/* eslint-disable */

import * as d3 from 'd3';
import { globalSetting } from '../global';
import { NetlistRender } from '.';
import { LAYOUT_CONSTANT, LINE_WIDTH } from './layout';
import { svgResource } from '../skin/draw';
import { PulseLine } from '../skin/plusation';

export class WireRender {
    /**
     * 
     * @param {d3.Selection} selection
     * @param {NetlistRender} rootRender  
     */
    constructor(selection, rootRender) {
        this.selection = selection;
        this.rootRender = rootRender;
        
        /**
         * @type {BasicD3DataItem[]}
         */
        this.data = [];

        /**
         * @description id 到管理数据项的映射
         * @type {Map<string, BasicD3ManagmentItem[]>}
         */
        this.id2manager = rootRender.id2manager;

        this.idCounter = 0;

        this.arrowHeight = 12;
        this.arrowWidth = 12;
    }

    /**
     * @description 将 elknode 关于 wire 的数据添加为 d3 数据项目
     * @param {ElkPoint[]} points 
     * @param {ElkEdge} edge 
     */
    addAsD3DataItems(points, edge) {
        const linePaths = [];
        for (let i = 0; i < points.length; ++ i) {
            // 根据点的信息创建 path
            const command = i === 0 ? 'M': 'L';
            const x = points[i].x;
            const y = points[i].y;
            linePaths.push(`${command}${x},${y}`);
        }

        const lineSvg = linePaths.join(' ');

        // 判断当前的朝向
        const endPoint = points.at(-1);
        const direction = endPoint.x > points.at(-2).x ? 'right' : 'left';
        const arrowX = direction === 'right' ? endPoint.x - LAYOUT_CONSTANT.CELL_PORT_WIDTH - this.arrowWidth:
                                               endPoint.x + LAYOUT_CONSTANT.CELL_PORT_WIDTH
        
        const arrowY = endPoint.y - this.arrowHeight / 2;
        

        this.data.push({
            id: this.idCounter,
            svg: lineSvg,
            endPoint: points.at(-1),
            arrow: {
                icon: direction + '-arrow',
                x: arrowX,
                y: arrowY,
                width: this.arrowWidth,
                height: this.arrowHeight
            },
            active: false
        });

        this.idCounter ++;
    }

    render() {
        const data = this.data;
        const id2manager = this.id2manager;
        const _this = this;

        const arrowHeight = 12;
        const arrowWidth = 12;


        let lineSelections = this.selection.selectAll('path.lines')
            .data(data)
            .enter()
            .append('path')
            .attr('d', d => d.svg)
            .attr('class', 'connection-line')
            .attr("fill", "none")
            .attr('stroke', 'var(--wire-color)');
        
        const arrows = new Map();

        let arrowSelections = this.selection.selectAll('svg.line-arrow')
            .data(data)
            .enter()
            .append(data => {
                const arrowInfo = data.arrow;

                // 获取箭头 svg
                const element = svgResource.get(arrowInfo.icon);

                element.setAttribute('x', arrowInfo.x);
                element.setAttribute('y', arrowInfo.y);
                element.setAttribute('width', arrowInfo.width);
                element.setAttribute('height', arrowInfo.height);
                element.setAttribute('opacity', 'var(--line-arrow-opacity)');
                
                return element;
            })
            .each(function(data) {
                const selection = d3.select(this);
                arrows.set(data.id, selection);
            });

        arrowSelections.raise();

        const id2animation = new Map();
        const id2PluseLine = new Map();

        lineSelections.on('click', function(_, data) {
            data.active = !data.active;

            if (data.active) {
                const pathSelection = d3.select(this);
                // 如果当前激活，显示数据流向
                // const pathLength = pathSelection.node().getTotalLength();

                const g = _this.selection.append("g")
                .attr("mask", "url(#m1)");
              
              // 创建 mask
              const mask = g.append("mask")
                .attr('id', 'm1')
              
              mask.append("use")
                .attr("id", "u1")
                .attr("href", "#p1")
                .attr("stroke", "white")
                .attr("stroke-width", 7)
                .attr("fill", "none")
                .attr('stroke-dasharray', '200 300')
                .attr("stroke-linecap", "round");
              
              // 创建 radialGradient
              const radialGradient = g.append("radialGradient")
                .attr("id", "g1");
              
              radialGradient.append("stop")
                .attr("offset", "0%")
                .attr("stop-color", "white");
              
              radialGradient.append("stop")
                .attr("offset", "25%")
                .attr("stop-color", "#CB81DA");
              
              radialGradient.append("stop")
                .attr("offset", "95%")
                .attr("stop-opacity", 0)
                .attr("stop-color", "rgb(70, 70, 222)");
              
              // 创建 path
              g.append("path")
                .attr("id", "p1")
                .attr('stroke-width', 10)
                .attr("fill", "none")
                .attr("d", data.svg);
              
              // 创建 circle 和 animateMotion
              g.append("circle")
                .attr("r", 100)
                .attr("fill", "url(#g1)")
                .append("animateMotion")
                .attr("dur", "3000ms")
                .attr("repeatCount", "indefinite")
                .append("mpath")
                .attr("href", "#p1");

                let p1 = document.getElementById('p1');
                let u1 = document.getElementById('u1');

                const pathLength = p1.getTotalLength();
                
                var u1Keyframes = new KeyframeEffect(
                  u1, [{
                      strokeDasharray: `200 ${pathLength - 200}`,
                      strokeDashoffset: "200"
                    },
                    {
                      strokeDasharray: `200 ${pathLength}`,
                      strokeDashoffset: `-${pathLength - 200}`
                    }
                  ], {
                    duration: 3000,
                    iterations: Infinity
                  }
                );
                
                const pulsationAnimation = new Animation(u1Keyframes, document.timeline);
                
                requestAnimationFrame(() => {
                    const motion = g.selectAll('animateMotion').node();
                    console.log(motion);
                    motion.setAttribute('begin', '0s');
                    motion.beginElement();
                    
                    pulsationAnimation.play();
                });

                // const pivot = _this.selection
                //     .append('circle')
                //     .attr('r', 6)
                //     .attr('fill', 'var(--wire-ball-color)');
                
                // // 进行一次动画
                // function renderOneAnimation() {
                //     console.log(pathLength);
                //     // 1400 的长度差不多配 3500 ?
                //     const duration = pathLength / 14 * 35;
                //     let transition = pivot.transition().duration(3500);
                //     transition.attrTween('transform', () => {
                //         return t => {
                //             const x = cubicBezierAnimation(t, 0, pathLength);
                //             const point = pathSelection.node().getPointAtLength(x);
                //             return `translate(${point.x},${point.y})`;
                //         }
                //     });
                //     transition = transition.on('end', function() {
                //         if (id2animation.has(data.id)) {
                //             renderOneAnimation();
                //         }
                //     });
                //     // 动画结束后重新启动动画
                //     id2animation.set(data.id, {
                //         pivot,
                //         transition
                //     });
                // }

                // renderOneAnimation();
            } else {
                if (id2animation.has(data.id)) {
                    const { pivot, transition } = id2animation.get(data.id);
                    pivot.remove();
                    id2animation.delete(data.id);
                }
            }
        });

        lineSelections.on('mouseenter', function(_, data) {
            const selection = d3.select(this);
            selection.raise();
            selection.attr('stroke', 'var(--wire-active-color)');

            const arrowSelection = arrows.get(data.id);            
            arrowSelection.raise();
            arrowSelection.selectAll('path').attr('fill', 'var(--wire-active-color)');
        });

        lineSelections.on('mouseleave', function(_, data) {
            const selection = d3.select(this);
            selection.attr('stroke', 'var(--wire-color)');

            const arrowSelection = arrows.get(data.id);
            arrowSelection.selectAll('path').attr('fill', 'var(--wire-color)');
        });

        
        if (globalSetting.renderAnimation) {
            lineSelections = lineSelections
                .transition()
                .duration(1000);
        }

        lineSelections
            .attr('stroke-width', LINE_WIDTH)
            .each(function (data) {
                const selection = d3.select(this);
                // const manager = _this.createDataManager(selection, data);
            });
    }

    /**
     * 
     * @param {d3.Selection} selection 
     * @param {BasicD3DataItem} data 
     * @returns {BasicD3ManagmentItem}
     */
    createDataManager(selection, data) {
        const id2manager = this.id2manager;
        // wire 不需要拖拽上下文
        const managerItem = {
            data,
            selection,
            type: 'wire'
        };

        if (!id2manager.has(data.id)) {
            id2manager.set(data.id, []);
        }

        id2manager.get(data.id).push(managerItem);
        return managerItem;
    }
}

/**
 * @description 计算贝塞尔插值，输入 0 - 1 的值 (delta)，输出 [oldVal, newVal] 中间的插值
 * @param {*} delta 
 * @param {*} oldVal 
 * @param {*} newVal 
 * @returns 
 */
function cubicBezierAnimation(delta, oldVal, newVal) {
    delta = 3 * (1 - delta) * (1 - delta) * delta + 3 * (1 - delta) * delta * delta + delta * delta * delta;
    return (1 - delta) * oldVal + delta * newVal;
}